// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_defender_move.h"

#include "bhv_basic_tackle.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "original/body_zulu.h"
#include "strategy.h"


/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_DefenderMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_DefenderMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    
    ////////////zulu////////////
    if( Body_Zulu().hear(agent) ){
               agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
               return true;
    }
    
    // tackle

    if ( Bhv_BasicTackle( 0.9, 80.0 ).execute( agent ) )
    {
        return true;
    }


    // kickableOpponentȼδ֤˼ʬ뤫
    bool can_block = false;
    // kickableOpponentȼδ֤Ũ뤫
    bool exist_opp = false;
    if( wm.getOpponentNearestToBall(5) ){
        rcsc::Triangle2D
            tri(wm.getOpponentNearestToBall(5)->pos(),
                rcsc::Vector2D( -rcsc::ServerParam::i().pitchHalfLength(),
                                rcsc::ServerParam::i().pitchHalfWidth() ),
                rcsc::Vector2D( -rcsc::ServerParam::i().pitchHalfLength(),
                                -rcsc::ServerParam::i().pitchHalfWidth() ) );
        
        if( tri.contains( wm.self().pos() ) ) can_block = true;
        const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
        const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
        for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
              it != opps_end;
              ++it )
        {
            if( tri.contains( (*it)->pos() ) ){
                exist_opp = true;
                break;
            }
        }
    }

    // intercept
    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();
    if ( ! wm.existKickableTeammate()
         && ( self_min <= 3
              || ( self_min < mate_min + 2
                   && self_min < opp_min + 3 ) )
         // ǥե󥹥饤ȴƤȤ
         // 饤󥿡ץȤʤ褦
         && ! ( wm.self().pos().x > -36.0
                && wm.self().pos().x < 0
                && wm.existKickableOpponent()
                && can_block )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: intercept self_min=%d mate_min=%d opp_min=%d"
                            ,__FILE__, __LINE__,
                            self_min,mate_min,opp_min);
        rcsc::Body_Intercept().execute( agent );
        
        if ( wm.ball().distFromSelf()
             < rcsc::ServerParam::i().visibleDistance() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );;
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }


    const double dash_power = getDashPower( agent, M_home_pos );
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;


    //block our goal
    if( wm.self().pos().x < -40.0
       && (wm.self().unum() == 4 || wm.self().unum() == 5) ){

        bool intercept_flg = false;
        rcsc::Vector2D v(0,0);
        const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
        const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
        for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
              it != opps_end;
              ++it )
        {
            if( (*it)->pos().x > -36
                && (*it)->pos().absY() > 8 ) continue;

            if( (*it)->pos().y * wm.self().pos().y < 0) continue;

            if( (*it)->pos().x < v.x &&
                (*it)->pos().absY() > v.absY() ){
                v = (*it)->pos();
                intercept_flg = true;
            }
        }

        rcsc::Vector2D v_to_ball = wm.ball().pos() - v;
        rcsc::Vector2D target_point = v + v_to_ball.setLengthVector( 2.0 ) + rcsc::Vector2D(-1.5,0);

        if( intercept_flg ){
            agent->debugClient().addMessage( "hara" );
            agent->debugClient().setTarget( target_point );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: hara goal"
                                ,__FILE__, __LINE__ );
            
            if ( ! rcsc::Body_GoToPoint( target_point, 0.5, dash_power
                     ).execute( agent ) )
            {
                rcsc::Body_TurnToBall().execute( agent );
            }
            
            if ( wm.existKickableOpponent()
                 && wm.ball().distFromSelf() < 18.0 )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            }
            
            return true;
        }
        
        
        
         bool block_flg = false;
         rcsc::Vector2D block_point( -49.0,
                                     rcsc::ServerParam::i().goalHalfWidth() - 3.0 );
         if( wm.ball().pos().y * M_home_pos.y < 0
            && ! wm.existKickableTeammate()
            && wm.ball().pos().x < -40
            && 5 < wm.ball().pos().absY() ){
            block_point.y = 0.5;
            block_flg = true;
        }
        if( wm.ball().pos().y * M_home_pos.y < 0
            && wm.existKickableOpponent()
            && wm.ball().pos().x < -40
            && 6 < wm.ball().pos().absY()
            && wm.ball().pos().absY() < 25 ){
            block_point.y = rcsc::ServerParam::i().goalHalfWidth() - 1.5;
            block_flg = true;
        }
        if( M_home_pos.y < 0 ) block_point.y *= -1.0;

        if(block_flg){
            agent->debugClient().addMessage( "block" );
            agent->debugClient().setTarget( block_point );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: block our goal"
                                ,__FILE__, __LINE__ );
            
            if ( ! rcsc::Body_GoToPoint( block_point, dist_thr, dash_power
                     ).execute( agent ) )
            {
                rcsc::Body_TurnToBall().execute( agent );
            }
            
            if ( wm.existKickableOpponent()
                 && wm.ball().distFromSelf() < 18.0 )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            }
            
            return true;
        }
    }
    
    //normal move
    rcsc::Vector2D new_home_pos = M_home_pos;
        
    //nearest_DFõ
    int nearest_DF = 0;
    if( ! wm.existKickableTeammate() )
    {
        double min_dist = 100, tmp_dist;
        // find DF nearest to ball
        for(int i=2;i<=5;i++){
            if(  wm.teammate(i) != static_cast< rcsc::PlayerObject * >( 0 ) ){
                tmp_dist = wm.teammate(i)->pos().dist( wm.ball().pos() );
                if( tmp_dist < min_dist
                    && wm.teammate(i)->pos().x < wm.ball().pos().x + 1 ){
                    min_dist = tmp_dist;
                    nearest_DF = i;
                }
            }
        }
        
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: nearestDF = %d"
                            ,__FILE__, __LINE__, nearest_DF );

        //--롼ѥк˥ǥե󥹥饤󲼤
        if( ( nearest_DF != wm.self().unum() || exist_opp )
            && Strategy::get_ball_area( M_home_pos ) != Strategy::BA_Danger )
        {
            if( M_home_pos.x < 0 && M_home_pos.x > -20 )
                new_home_pos.x -= -5 * (new_home_pos.x)/20.0;
            if( M_home_pos.x <= -20 )
                new_home_pos.x -= 5;
        }

        //--ʬֶᤫäɥ֥˸֤˹Ԥ
        // ŨˤȤϤʤ褦
        if( nearest_DF == wm.self().unum()
            && Strategy::get_ball_area( wm.ball().pos() ) != Strategy::BA_Danger
            && wm.ball().pos().x < wm.self().pos().x + 7
            && wm.ball().pos().x < wm.defenseLineX() + 10
            && ! exist_opp )
        {
            rcsc::AngleDeg block_angle
                = (rcsc::Vector2D(-60,0) - wm.ball().pos()).th();
            
            if( wm.self().pos().absY() < 6 ){
                new_home_pos
                    = wm.ball().pos()
                    + rcsc::Vector2D(-5.0,0);
            } else if( wm.self().pos().x < -45 ){
                //첿
                if( wm.self().pos().y >= 0 ){
                    new_home_pos
                        = wm.ball().pos()
                        + rcsc::Vector2D(0,-3.0);
                } else {
                    new_home_pos
                        = wm.ball().pos()
                        + rcsc::Vector2D(0,3.0);
                }
            } else {
                new_home_pos
                    = wm.ball().pos()
                    + rcsc::Vector2D::polar2vector(5.0,block_angle);
            }
            agent->debugClient().addMessage( "blockD" );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: block dribble"
                                ,__FILE__, __LINE__);
        }
    }
        
    agent->debugClient().addMessage( " BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( new_home_pos );
    
    if ( ! rcsc::Body_GoToPoint( new_home_pos, dist_thr, dash_power
             ).execute( agent ) )
    {
        rcsc::Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
    

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Bhv_DefenderMove::getDashPower( const rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & target_point )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    const double stamina_max = rcsc::ServerParam::i().staminaMax();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = rcsc::ServerParam::i().maxDashPower();
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();


    /*------defender-----*/
    if( wm.ball().pos().x < wm.self().pos().x-0.5 )
        return rcsc::ServerParam::i().maxPower();
    
    if( wm.ball().pos().x >= -35.0 && wm.existKickableTeammate()
        && wm.self().stamina() <= stamina_max * 0.875 )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;
        return dash_power;
    }
    
    else if( wm.ball().pos().x <= -15.0 ){
        if( wm.ball().pos().x <= wm.defenseLineX() + 10 )
            return rcsc::ServerParam::i().maxPower();
        else
            return std::min( my_inc * 1.7,
                             rcsc::ServerParam::i().maxPower() );
    }
    
    else if(wm.ball().pos().x >= 0.0){
        if(wm.self().stamina() <= stamina_max * 0.875 ){
            //rest
            return std::max( 0.0, my_inc - 25.0 );
        } else {
            dash_power = std::min( my_inc * 1.5,
                                   rcsc::ServerParam::i().maxPower() );
            return dash_power;
        }
        
    } else {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        return dash_power;
    }
    /*-------------------*/


    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": correct DF line. keep max power" );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxDashPower();
    }
    else if ( ! wm.existKickableTeammate()
              && opp_min < self_min
              && opp_min < mate_min
              && target_point.x < wm.self().pos().x
              && target_point.x < -25.0
              && wm.ball().inertiaPoint( opp_min ).x < -25.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": defense dash" );
        dash_power = rcsc::ServerParam::i().maxDashPower();
    }

    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": recovering" );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": exist kickable teammate. dash_power=%.1f",
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": in offside area. dash_power=%.1f",
                            dash_power );
    }
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": normal mode dash_power=%.1f",
                            dash_power );
    }

    return dash_power;
}
