// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_midfielder_move.h"

#include "bhv_basic_tackle.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/geom/polygon_2d.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "original/body_zulu.h"
#include <rcsc/action/intention_time_limit_action.h>
#include "strategy.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_MidfielderMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_MidfielderMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    
    //zulu
    if( Body_Zulu().hear(agent) ){
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return true;
    }

    // tackle
    if (! wm.existKickableTeammate() &&
        Bhv_BasicTackle( 0.9, 80.0 ).execute( agent ) )
    {
        return true;
    }

    // intercept
    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( ! wm.existKickableTeammate()  //magic number
         && ( self_min <= 3
              || ( self_min < mate_min + 3
                   && self_min < opp_min + 4 )
             )
        )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        if ( wm.ball().distFromSelf()
             < rcsc::ServerParam::i().visibleDistance() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );;
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }

    //midfielder intercept
     double min_dist = 100,tmp_dist;
     double nearest_MF = 0;
     const double stamina_max = rcsc::ServerParam::i().staminaMax();

//     if( ! wm.existKickableTeammate() )
//     {
        // find MF nearest to ball
     for(int i=6;i<=8;i++){
         if(  wm.teammate(i) != static_cast< rcsc::PlayerObject * >( 0 ) ){
             tmp_dist = wm.teammate(i)->pos().dist( wm.ball().pos() );
             if( tmp_dist < min_dist ){
                 min_dist = tmp_dist;
                 nearest_MF = i;
             }
         }
     }
     
     //ץܡ뤬ɤˤư
     if( wm.existKickableOpponent() &&
         wm.ball().pos().absY() > 20 ){
         if( nearest_MF == wm.self().unum() ){
            
             if ( ! rcsc::Body_GoToPoint( wm.ball().pos() + rcsc::Vector2D( 2, wm.self().pos().y > 0 ? -1.5 : 1.5 ),
                                          0.5,
                                          rcsc::ServerParam::i().maxPower()
                      ).execute( agent ) )
             {
                 rcsc::Body_Intercept().execute( agent );
             }
             
             if ( wm.existKickableOpponent()
                  && wm.ball().distFromSelf() < 18.0 )
             {
                 agent->setNeckAction( new rcsc::Neck_TurnToBall() );
             }
             else
             {
                 agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
             }
             return true;
         }
         else if( wm.self().unum() == 6 )
         {
             rcsc::Vector2D self_to_ball = (wm.ball().pos() - M_home_pos) * 0.8;
             rcsc::Vector2D side = self_to_ball.rotatedVector(90).setLengthVector(7.0);
             rcsc::Vector2D self_back = M_home_pos - self_to_ball.setLengthVector(5);
             std::vector< rcsc::Vector2D > v;
             v.push_back( self_back + side );
             v.push_back( self_back + self_to_ball + side );
             v.push_back( self_back + self_to_ball - side );
             v.push_back( self_back - side );
             rcsc::Polygon2D search_region(v);
             
             agent->debugClient().addLine( self_back - side,
                                           self_back + side );
             agent->debugClient().addLine( self_back + side,
                                           M_home_pos + self_to_ball + side );
             agent->debugClient().addLine( M_home_pos + self_to_ball + side,
                                           M_home_pos + self_to_ball - side );
             agent->debugClient().addLine( M_home_pos + self_to_ball - side,
                                           self_back - side );

             double front_x = 50;
             rcsc::Vector2D target;
             const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
             const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
             for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
                   it != opps_end;
                   ++it )
             {
                 if( search_region.contains( (*it)->pos() ) )
                 {
                     if( front_x > (*it)->pos().x ){
                         front_x = (*it)->pos().x;
                         rcsc::Vector2D it_to_ball = wm.ball().pos() - (*it)->pos();
                         target = (*it)->pos() + it_to_ball * 0.3;
                     }
                 }
             }
             agent->debugClient().setTarget( target );

             
//             agent->setIntention(
//                 new rcsc::IntentionTimeLimitAction( new rcsc::Body_GoToPoint( wm.ball().pos() + rcsc::Vector2D( 2, wm.self().pos().y > 0 ? -1.5 : 1.5 ),
//                                           0.5, rcsc::ServerParam::i().maxPower() ),
//                                                     new rcsc::Neck_TurnToBall(),
//                                                     NULL,
//                                                     5,
//                                                     wm.time() ) );
            
             if ( ! rcsc::Body_GoToPoint( target, 0.5,
                                          rcsc::ServerParam::i().maxPower() * 0.9
                      ).execute( agent ) )
             {
                 rcsc::Body_TurnToBall().execute( agent );
             }
                
            
             if ( wm.existKickableOpponent()
                  && wm.ball().distFromSelf() < 18.0 )
             {
                 agent->setNeckAction( new rcsc::Neck_TurnToBall() );
             }
             else
             {
                 agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
             }
             return true;
             
         }
     }
        
        
//         bool flg = false;
//         static bool S_recovery = false;
//         if( wm.ball().pos().x > -20 && wm.ball().pos().x < 20
//             && nearest_MF == wm.self().unum() ){
//             switch( wm.self().unum() ){
//             case 6:
//                 if( wm.ball().pos().y > -15 && wm.ball().pos().y < 15 )
//                     flg = true;
//                 break;
//             case 7:
//                 if( wm.ball().pos().y < -5 )
//                     flg = true;
//                 break;
//             case 8:
//                 if( wm.ball().pos().y > 5 )
//                     flg = true;
//                 break;
//             default:
//                 break;
//             }
//         }

//         const rcsc::PlayerObject * fastest_opp
//             = wm.interceptTable()->fastestOpponent();

//         if( flg && fastest_opp ){
//             rcsc::Vector2D block_point = fastest_opp->pos();
//             if( wm.self().pos().x > fastest_opp->pos().x ){
//                 block_point.x -= 5.0;
//             } else {
//                 block_point.x -= 2.0;
//             }

//             //magic number
//             if( wm.self().stamina() >= stamina_max * 0.625 )
//                 S_recovery = false;

//             if(!S_recovery && wm.self().stamina() >= stamina_max * 0.5 ){
//                 if ( rcsc::Body_GoToPoint( block_point, 1.0,
//                                            rcsc::ServerParam::i().maxPower(),
//                                            5,false,true,40.0).execute( agent ) );
//                 {
//                     rcsc::dlog.addText( rcsc::Logger::TEAM,
//                                         "%s:%d: MF intercept"
//                                         ,__FILE__, __LINE__ );
//                     if ( wm.ball().distFromSelf()
//                          < rcsc::ServerParam::i().visibleDistance() )
//                     {
//                         agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );;
//                     }
//                     else
//                     {
//                         agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
//                     }
//                     return true;
//                 }
//             }
//             else
//             {
//                 S_recovery = true;
//             }
//         }
//     }


    // midfielder move(danger)
    //--褦
    if( wm.existKickableOpponent()
        && wm.ball().pos().x < -32 && wm.ball().pos().absY() < 20 
        && nearest_MF == wm.self().unum() )
    {
        rcsc::Vector2D center(-60,0);
        rcsc::Vector2D block_point = wm.ball().pos();
        if( wm.ball().pos().x < -45 ){
            block_point
                += rcsc::Vector2D::polar2vector(2,
                                                (center-wm.ball().pos()).th() );
        } else {
            block_point
                += rcsc::Vector2D::polar2vector(5,
                                                (center-wm.ball().pos()).th() );
        }
        
        agent->debugClient().addMessage( " MF danger");
        agent->debugClient().setTarget( block_point );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: MF danger dist=%f"
                            ,__FILE__, __LINE__,
                            wm.self().pos().dist(block_point) );

        if( wm.self().pos().dist(block_point) < 1.5 ){  //magic number
            agent->setIntention(
                new rcsc::IntentionTimeLimitAction( new rcsc::Body_Intercept(),
                                                    new rcsc::Neck_TurnToBall(),
                                                    NULL,
                                                    10,
                                                    wm.time() ) );
            rcsc::Body_Intercept().execute(agent);
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }

        if ( ! rcsc::Body_GoToPoint( block_point, 0.5,
                                     rcsc::ServerParam::i().maxPower()
                 ).execute( agent ) )
        {
            rcsc::Body_TurnToBall().execute( agent );
        }
                
        if ( wm.existKickableOpponent()
             && wm.ball().distFromSelf() < 18.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }


    const double dash_power = getDashPower( agent, M_home_pos );
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;
    rcsc::Vector2D new_home_pos = M_home_pos;

    //zuluѤhome_posĴ
     double offside = wm.offsideLineX();
//     const double GO = 0;
//     const double ARRIVE = 10;
//     if( wm.ball().pos().x > GO && wm.ball().pos().x < ARRIVE ){
//         new_home_pos.x +=  (offside - new_home_pos.x)
//             * (new_home_pos.x - GO)
//             / (ARRIVE - GO);
//         double mates_dist=0;
//     }
//     if( wm.ball().pos().x >= ARRIVE ){
//         new_home_pos.x = offside;
//     }
//     if( wm.ball().pos().x > 32 )
//         new_home_pos = M_home_pos;

    
    //normal move
     
    //--offenseŨΥ
    if( wm.existKickableTeammate()
        && wm.ball().pos().absX() < 25
        && wm.getDistOpponentNearestToSelf(5) < 3 )
    {
        const rcsc::PlayerObject * near_opp = wm.getOpponentNearestToSelf(5);
        if( near_opp != static_cast< rcsc::PlayerObject * >( 0 ) ){
            new_home_pos.x = std::min( near_opp->pos().x + 5,
                                       offside );
        }
        agent->debugClient().addMessage( "avoid opp");
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: avoid near opponent"
                            ,__FILE__, __LINE__ );
    }
    
    //--꤬֤ʤ褦
    if( Strategy::get_ball_area( wm.ball().pos() ) != Strategy::BA_Danger
        && wm.existKickableTeammate()
        && new_home_pos.dist( wm.ball().pos() ) <= 5.0 )
    {
        if( wm.self().pos().x > wm.ball().pos().x )
            new_home_pos.x = std::min(new_home_pos.x+5.0,offside);
        else
            new_home_pos.x = std::min(new_home_pos.x-5.0,offside);
    }

    agent->debugClient().addMessage( " BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( new_home_pos );
        
    if ( ! rcsc::Body_GoToPoint( new_home_pos, dist_thr, dash_power
             ).execute( agent ) )
    {
        rcsc::Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Bhv_MidfielderMove::getDashPower( const rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & target_point )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = rcsc::ServerParam::i().maxDashPower();
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();
    
    if( wm.existKickableOpponent() && wm.ball().pos().x <= -15.0 ){
        if( wm.ball().pos().x <= wm.defenseLineX() + 10 ){
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: danger. keep max power"
                                ,__FILE__, __LINE__ );
            return rcsc::ServerParam::i().maxPower();
        } 
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: defense. keep max power"
                                ,__FILE__, __LINE__ );
            return std::min( my_inc * 1.7,
                             rcsc::ServerParam::i().maxPower() );
        }
    }

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": correct DF line. keep max power" );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxDashPower();
    }
    else if ( ! wm.existKickableTeammate()
              && opp_min < self_min
              && opp_min < mate_min
              && target_point.x < wm.self().pos().x
              && target_point.x < -25.0
              && wm.ball().inertiaPoint( opp_min ).x < -25.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": defense dash" );
        dash_power = rcsc::ServerParam::i().maxDashPower();
    }

    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": recovering" );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": exist kickable teammate. dash_power=%.1f",
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": in offside area. dash_power=%.1f",
                            dash_power );
    }
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": normal mode dash_power=%.1f",
                            dash_power );
    }

    return dash_power;
}
