#include "body_kubo_dribble.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_dribble.h>
#include <rcsc/geom/polygon_2d.h>

#include "original/offense_system.h"
#include <rcsc/action/neck_scan_field.h>

#include <iostream>

#include <rcsc/action/body_kick_one_step.h>
#include "original/intention_dribble.h"


//ȾM_distȾߤ˱äãͽ¬
//M_afraid;͵äɥ֥
bool
Body_KuboDribble::execute(rcsc::PlayerAgent * agent)
{
    double dist = M_dist;
    int kubo_afraid = M_afraid;
    
    rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                        "%s:%d: !start kubodribble!"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    bool flg=true;
    std::vector< rcsc::Vector2D > kubo_point;

    //dir from body to goal
    rcsc::AngleDeg want_dir = ( rcsc::Vector2D(50.0,0.0) - wm.self().pos() ).th();
    //set kubo_point
    rcsc::Vector2D tmp(1.0,0.0);
    tmp.setLength(dist);
    tmp.setDir( rcsc::AngleDeg( want_dir - 90));

    for(int i=0;i<13;i++){
        kubo_point.push_back( wm.self().pos() + tmp );

        if(kubo_point[i].x>(rcsc::ServerParam::i().pitchHalfLength()-2))
            kubo_point[i].x=rcsc::ServerParam::i().pitchHalfLength()-2;
        if(kubo_point[i].y<(-rcsc::ServerParam::i().pitchHalfWidth()+2))
            kubo_point[i].y=-rcsc::ServerParam::i().pitchHalfWidth()+2;
        if(kubo_point[i].y>(rcsc::ServerParam::i().pitchHalfWidth()-2))
            kubo_point[i].y=rcsc::ServerParam::i().pitchHalfWidth()-2;

        tmp.rotate(15);
    }

    //search safety direction
    int direction;
    flg = searchSafetyDir( agent,kubo_point,direction,kubo_afraid);

    //dribble or holdBall
    if(flg){
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                            "%s:%d: !dribble to %d!"
                            ,__FILE__, __LINE__, direction );
        agent->debugClient().addLine
            ( wm.self().pos(),kubo_point[direction]);
	 
//         rcsc::Body_Dribble(kubo_point[direction],
//                            0.5,
//                            rcsc::ServerParam::i().maxPower(),
//                            1,
//                            false).execute(agent);

        agent->setIntention
            ( new IntentionDribble( kubo_point[direction],
                                    1.0,
                                    5,
                                    wm.time() ) );
        
    double max_vel
        = rcsc::Body_KickOneStep::get_max_possible_vel( kubo_point[direction].th(),
                                                        wm.self().kickRate(),
                                                        wm.ball().vel() ).r();

    //ǥɥ֥δĴ
    const double dribble_dist = 5.0;
    const double ball_reach_cycle = 10;
    
    double first_speed = rcsc::calc_first_term_geom_series( dribble_dist,
                                                         rcsc::ServerParam::i().ballDecay(),
                                                         ball_reach_cycle );

    if( max_vel < first_speed )
        first_speed = max_vel;

    agent->debugClient().setTarget( kubo_point[direction] );
    rcsc::Body_KickOneStep( kubo_point[direction], first_speed, true ).execute( agent );
        
    } else {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                            "%s:%d: !can't dribble!"
                            ,__FILE__, __LINE__ );
    }  

    return flg;
}

bool
Body_KuboDribble::searchSafetyDir(rcsc::PlayerAgent * agent,
                                  const std::vector< rcsc::Vector2D > & kubo_point,
                                  int &direction,
                                  const int kubo_afraid)
{
  int add,goal_dir=6;
  const rcsc::WorldModel & wm = agent->world();

  //no opp
  if( agent->world().opponents().empty() ){
	 rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								"%s:%d: !opponent is empty!"
								,__FILE__, __LINE__ );
	
	 direction = goal_dir;
	 return true;
  }

  int self_cycle;
  int opp_cycle;
  bool flg;
  
  //direction loop
  for(int i=0;i<=6;i++){

	 add = i;

	 //---check one side---	
	 flg = true;
	 if( wm.self().pos().y > 0.0 && add != 0)add *= -1;
	 direction = goal_dir + add;

	 rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								"%s:%d: !search dir= %d!"
								,__FILE__, __LINE__, direction );

	 self_cycle = OffenseSystem::getPointReachCycle(agent,
                                             kubo_point[direction],
                                             wm.self());
	 opp_cycle = OffenseSystem::pointPredict(agent,kubo_point[direction],
                                      10,OPP,NULL,NULL );

	 if(self_cycle != -1 ||
		 opp_cycle != -1 || 
		 opp_cycle - self_cycle < kubo_afraid){
		flg = false;
		rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								  "%s:%d: !afraid diff= %d, afraid= %d!"
								  ,__FILE__, __LINE__,
								  opp_cycle-self_cycle, kubo_afraid);
	 }
	 if(flg)break;
	

	 //---skip if direction==goal_dir---
	 if(add == 0)continue;

	
	 //---check other side---
	 add *= -1;
	 direction = goal_dir + add;
	 flg = true;

	 rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								"%s:%d: !search dir= %d!"
								,__FILE__, __LINE__, direction );

	 self_cycle = OffenseSystem::getPointReachCycle(agent,
                                             kubo_point[direction],
                                             wm.self());
     OffenseSystem::pointPredict(agent,kubo_point[direction],
                          10,OPP,&opp_cycle,NULL );

	 if(self_cycle <= 0 || 
		 opp_cycle <= 0 ||
		 opp_cycle - self_cycle < kubo_afraid){
		flg = false;
		rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								  "%s:%d: !afraid diff= %d, afraid= %d!"
								  ,__FILE__, __LINE__,
								  opp_cycle-self_cycle, kubo_afraid);
	 }
	 if(flg) break;
  }

  return flg;


    return true;
}
