#include "body_one_on_one_dribble.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_dribble.h>
#include <rcsc/geom/polygon_2d.h>

#include "offense_system.h"
#include <rcsc/action/body_kick_one_step.h>
#include "intention_dribble.h"

#include <vector>

bool
Body_OneOnOneDribble::execute(rcsc::PlayerAgent * agent)
{
    rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                        "%s:%d: !one on one start!"
                        ,__FILE__, __LINE__ );
    
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();

    if( wm.self().pos().x < 0.0) return false;

    //get goalie info;
    double goalie_pos_x = 46.0;
    const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
		  it != opps_end;
		  ++it )
    {
		if( (*it)->goalie() )
        {
            goalie_pos_x = (*it)->pos().x - 5.0;
            break;
        }
    }

    //check cycle
    int self_cycle,opp_cycle,max_diff=-100;
    int best_idx=0;
    rcsc::Vector2D shoot_pos_candidate[5];
    for(int i=0;i<5;i++){

        if( wm.self().pos().y > 0 )
            shoot_pos_candidate[i].assign( goalie_pos_x, -6.0+(3.0*(double)i) );
        else
            shoot_pos_candidate[i].assign( goalie_pos_x, 6.0-(3.0*(double)i) );


        self_cycle = OffenseSystem::getPointReachCycle( agent,
                                                 shoot_pos_candidate[i],
                                                 wm.self());
        
        OffenseSystem::pointPredict( agent, shoot_pos_candidate[i],10,
                                     OPP, &opp_cycle, NULL,true);
        
        if( opp_cycle - self_cycle >= max_diff ){
            max_diff = opp_cycle - self_cycle;
            best_idx = i;
        }
    }
    rcsc::Vector2D shoot_pos = shoot_pos_candidate[best_idx];

    //checkOneOnOne
    bool free = true;
    bool goalie;
    std::vector< rcsc::Vector2D > points;
    if(  wm.self().pos().y > 0 ){
        points.push_back( rcsc::Vector2D( 52.0, 9.0));
        points.push_back( rcsc::Vector2D( 52.0, -30.0));
    } else {
        points.push_back( rcsc::Vector2D( 52.0, 30.0));
        points.push_back( rcsc::Vector2D( 52.0, -9.0) );
    }

    points.push_back( wm.self().pos() + rcsc::Vector2D(0,-0.1) );
    points.push_back( wm.self().pos() + rcsc::Vector2D(0,0.1) );
    const rcsc::Polygon2D poly( points );
    //opp loop
    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
		  it != opps_end;
		  ++it )
    {
        //goalieֹƤʤѽ
        if( wm.self().pos().x <= 40 && abs((*it)->pos().y) <= 9
            && (*it)->pos().x >= 47 ){
            goalie = true;
        } else {
            goalie = false;
        }
        
		if( !( (*it)->goalie() || goalie ) && poly.contains( (*it)->pos() ) )
        {
            free = false;
            break;
        }
    }
    //mate loop
    const rcsc::PlayerPtrCont & mates = wm.teammatesFromSelf();
    const rcsc::PlayerPtrCont::const_iterator mates_end = mates.end();
    for ( rcsc::PlayerPtrCont::const_iterator it = mates.begin();
		  it != mates_end;
		  ++it )
    {
        if( poly.contains( (*it)->pos() ) )
        {
            free = false;
            break;
        }
    }

    //dribble
    if(free){
        if( (wm.self().pos()-shoot_pos).r() <= 1.0 ){
            rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                                "%s:%d: !One on One already reached!"
                                ,__FILE__, __LINE__,wm.self().body().abs() );
            return false;
        }
            

        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                            "%s:%d: !One on One body=%f!"
                            ,__FILE__, __LINE__,wm.self().body().abs() );

        agent->debugClient().addLine( points[1],points[2] );
        agent->debugClient().addLine( points[2],points[3] );
        agent->debugClient().addLine( points[3],points[0] );
        agent->debugClient().addCircle( shoot_pos,2 );

        int dash_count=5;
        if( wm.self().pos().x > 41 )dash_count = 2;
        
//         static rcsc::Vector2D S_old_shoot_pos;
//         static int S_old_flg=false;
//         static int S_old_count=0;
        
//         if( wm.self().pos().x > wm.offsideLineX() - 5
//             && wm.self().pos().x < 36.0){
//             rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
//                                 "%s:%d: !One on One set Dribble Intention!"
//                                 ,__FILE__, __LINE__ );
//             S_old_flg = true;
//             S_old_shoot_pos = shoot_pos;
//         }

//         if( S_old_flg ){
//             S_old_count++;
//             rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
//                                 "%s:%d: !One on One Dribble Intention!"
//                                 ,__FILE__, __LINE__ );
//             rcsc::Body_Dribble( S_old_shoot_pos,
//                                 1.0,
//                                 rcsc::ServerParam::i().maxPower(),
//                                 dash_count,
//                                 false
//                 ).execute( agent );
//             if( S_old_count > 10 ){
//                 S_old_count = 0;
//                 S_old_flg = false;
//             }
//             return true;
//         }
        
//         rcsc::Body_Dribble( shoot_pos,
//                             1.0,
//                             rcsc::ServerParam::i().maxPower(),
//                             dash_count,
//                             false
//             ).execute( agent );

//         agent->setIntention
//             ( new IntentionDribble( shoot_pos,
//                                     1.0,
//                                     5,
//                                     wm.time() ) );
        
//         double max_vel
//             = rcsc::Body_KickOneStep::get_max_possible_vel( shoot_pos.th(),
//                                                             wm.self().kickRate(),
//                                                             wm.ball().vel() ).r();

//     //ǥɥ֥δĴ
//     const double dribble_dist = 5.0;
//     const double ball_reach_cycle = 7;
    
//     double first_speed = rcsc::calc_first_term_geom_series( dribble_dist,
//                                                             rcsc::ServerParam::i().ballDecay(),
//                                                             ball_reach_cycle );

//     if( max_vel < first_speed )
//         first_speed = max_vel;

//     agent->debugClient().setTarget( shoot_pos );
//     rcsc::Body_KickOneStep( shoot_pos, first_speed, true ).execute( agent );

        //offside lineۤƤʤ̤Υɥ֥
        if( wm.offsideLineX() > wm.self().pos().x ){
            rcsc::Body_Dribble( shoot_pos,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                dash_count, false ).execute( agent );
            

        } else if( wm.self().pos().x < 36 ){
            //®ɥ֥        
            rcsc::Body_KickOneStep( shoot_pos,
                                    1.0,
                                    true ).execute( agent );
            
            agent->setIntention
                ( new IntentionDribble( shoot_pos,
                                        1.0,
                                        8,
                                        wm.time() ) );
        }
        else
        {
            rcsc::Body_Dribble( rcsc::Vector2D(47,0),
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                dash_count, false ).execute( agent );
        }
        
        return true;
        
    }
    
    return false;
    
}
