#include "tile_sarsa.h"
#include <iostream>
#include <sstream>
#include <set>
#include <rcsc/random.h>
#include <algorithm>
#include <cstdlib>

//MaxValueõ
struct Value
{
    Value (const int & idx,
           const double & value )
    :M_idx(idx)
    ,M_value(value)
    { }
    
    int M_idx;
    double M_value;
    bool operator <(const Value & rhs)const { // 羮 less 黻
        return M_value < rhs.M_value;
    }
};

TileSarsa::TileSarsa( const double & epsilon,
                      const double & alpha,
                      const int & input_num,
                      const std::vector< std::vector<int> > & input_map,
                      const int & tiling_num,
                      const std::vector<double> & tile_widths )
 :M_epsilon(epsilon)
 ,M_alpha(alpha)
 ,M_input_num(input_num)
 ,M_input_map(input_map)
 ,M_tiling_num(tiling_num)
 ,M_tile_widths(tile_widths)
 ,M_tile_coding_num(input_map.size())
 ,M_episode_first_call(true)
 ,M_last_Q(0.0)
{
    //check error
    if( ( tile_widths.size() !=
          (unsigned)(input_num) ) ){
        std::cerr << "**ERROR** tile_widths size is not input_num(tile_sarsa.cpp)"
                  << std::endl;
    }

    //check error
    for(unsigned int i=0; i<input_map.size(); ++i){
        for(unsigned int j=0; j<input_map[i].size(); ++j){
            if( input_map[i][j] < 0 || input_map[i][j] >= input_num ){
                std::cerr << "**ERROR** invalid input_map(tile_sarsa.cpp)"
                          << std::endl;
            }
        }
    }

    //ACTION_NUM * tile_coding_numTileCoding饹
    tc.resize(ACTION_NUM);

    for(int i=0; i<ACTION_NUM; ++i){
        for(int j=0; j<M_tile_coding_num; ++j){
            //input_map˱ä׻
            std::vector<double> widths;
            for(unsigned int k=0; k<M_input_map[j].size(); ++k)
                widths.push_back( M_tile_widths[M_input_map[j][k]] );
            tc[i].push_back( TileCoding( input_map[j].size(),
                                         tiling_num,
                                         widths ) );
        }
    }
    
}

Action
TileSarsa::selectAction( const State & s )
{   
    //-greedy
    rcsc::UniformReal uni_real( 0, 1 );

    if( M_epsilon > uni_real() ){
        //---random action---
        //0ACTION_NUM-1(ACTION_NUM-1ޤ)
        rcsc::UniformInt uni_act( 0, ACTION_NUM-1 );
        return static_cast<Action>(uni_act());
    } else {
        //---greedy---
        //setȤäƥ
        std::multiset<Value> v;
        for(int i=0; i<ACTION_NUM; ++i){
            v.insert( Value(i,getValue(s,static_cast<Action>(i))) );
        }

        //ǥХå
//         std::set<Value>::reverse_iterator riit=v.rbegin();
//         while( riit != v.rend() )
//         {
//             cout << (*riit).M_idx << ":" << (*riit).M_value << " ";
//             ++riit;
//         }
//         cout << endl;
        

        //߽ߤΤreverse_iteratorȤ
        std::set<Value>::reverse_iterator rit=v.rbegin();
        std::set<Value>::reverse_iterator rit_before_end = --v.rend();
        int same_tail = 0;

        //ͤƱΤο
        while( rit != rit_before_end )
        {
            double tmp = (*rit).M_value;
            ++rit;
            if( std::abs(tmp - (*rit).M_value) < 10e-10 )
                ++same_tail;
            else
                break;
        }

        //ƱΤϥ
        rit=v.rbegin();
        if( same_tail > 0 ){
            rcsc::UniformInt uni_act( 0, same_tail );
            int target_idx = uni_act();
            for(int i=0; i<target_idx; ++i)
                ++rit;
        }
        
        return static_cast<Action>( (*rit).M_idx );
    }

}

bool
TileSarsa::update( const State & s, Action a, const double & r )
{
    
    if( M_episode_first_call ){
        M_last_state = s;
        M_last_action = a;
        M_last_Q = getValue( s,a );
        M_episode_first_call = false;
        return true;
    }

    double current_value = getValue( s,a );
    
//     std::cout << "r = " << r << " curQ = " << current_value
//               << " lastQ = " << M_last_Q
//               << " tderror = " <<  r + current_value - M_last_Q
//               << std::endl;
    
    //s,a,r,last_state,last_actionȤäQ
    updateValue( M_last_state, M_last_action,
                 r + current_value - M_last_Q );

    M_last_Q = current_value;
    M_last_state = s;
    M_last_action = a;


    return true;
}

bool
TileSarsa::update( const double & r )
{
    if( M_episode_first_call ){
        return true;
    }

//     std::cout << "r = " << r 
//               << " lastQ = " << M_last_Q 
//               << " tderror = " <<  r - M_last_Q
//               << std::endl;

    //r,last_state,last_actionȤäQ
    updateValue( M_last_state, M_last_action,
                 r - M_last_Q );

    return true;
}

bool
TileSarsa::readValue( const std::string & file_name )
{
    bool flg = true;
    for(int i=0; i<ACTION_NUM; ++i){
        for(int j=0; j<M_tile_coding_num; ++j){
            std::ostringstream oss;
            oss << file_name << "-" << i
                << "-" << j << ".dat";
            if( !tc[i][j].read( oss.str() ) ){
                flg = false;
            }
        }
    }

    return flg;
}

bool
TileSarsa::writeValue( const std::string & file_name )
{
    bool flg = true;
    for(int i=0; i<ACTION_NUM; ++i){
        for(int j=0; j<M_tile_coding_num; ++j){
            std::ostringstream oss;
            oss << file_name << "-" << i
                << "-" << j << ".dat";
            if( tc[i][j].write( oss.str() ) ){
                flg = false;
            }
        }
    }
    
    return flg;
}

double
TileSarsa::getValue( const State & s, Action a )
{
    std::vector<double> s_vec;

    s_vec = s.getStateList();

    int act = static_cast<Action>(a);
    double sum = 0;
    
    for(unsigned int i=0; i<tc[act].size(); ++i){
        //
        std::vector<double> input;
        for(unsigned int j=0; j<M_input_map[i].size(); ++j)
            input.push_back( s_vec[M_input_map[i][j]] );
        
        sum += tc[act][i].getValue(input);
    }

    //Ƴؽʬ䤷ƤʤΤǤǳؽǳ
    return sum/(double)tc[act].size();
}

bool
TileSarsa::updateValue( const State & s, Action a, const double & td_error )
{
    std::vector<double> s_vec;

    s_vec = s.getStateList();

    //td_error˳ؽΨ򤫤(ʬϤʤ)
//    static const double coef = M_alpha / (double)(M_tile_coding_num);
    static const double coef = M_alpha;
    double diff = coef * td_error;

    int act = static_cast<Action>(a);
    
    //input_map˽ätilecodingƽŤߤ¤
    for(unsigned int i=0; i<tc[act].size(); ++i){
        //
        std::vector<double> input;
        for(unsigned int j=0; j<M_input_map[i].size(); ++j)
            input.push_back( s_vec[M_input_map[i][j]] );

        tc[act][i].update( input, diff );
    }

    return true;
}

